#from .custom_numbers import CustomNumeralSystem, CustomNumber # Saving some typing to the users
from typing import List

__version__: str = "1.1.0"
__author__: str = r"Evgueni Antonov (Evgueni.Antonov@gmail.com)"
__all__: List[str] = ["custom_numbers"]
